"""bottom_bar.py"""
from typing import Any, Callable, Optional

from prompt_toolkit.formatted_text import HTML, merge_formatted_text
from prompt_toolkit.key_binding import KeyBindings
from rich.console import Console

from falyx.themes.colors import OneColors
from falyx.utils import CaseInsensitiveDict


class BottomBar:
    """Bottom Bar class for displaying a bottom bar in the terminal."""
    def __init__(self, columns: int = 3, key_bindings: KeyBindings | None = None):
        self.columns = columns
        self.console = Console()
        self._items: list[Callable[[], HTML]] = []
        self._named_items: dict[str, Callable[[], HTML]] = {}
        self._states: dict[str, Any] = CaseInsensitiveDict()
        self.toggles: list[str] = []
        self.key_bindings = key_bindings or KeyBindings()

    def get_space(self) -> int:
        return self.console.width // self.columns

    def add_static(
        self, name: str, text: str, fg: str = OneColors.BLACK, bg: str = OneColors.WHITE
    ) -> None:
        def render():
            return HTML(
                f"<style fg='{fg}' bg='{bg}'>{text:^{self.get_space()}}</style>"
            )

        self._add_named(name, render)

    def add_counter(
        self,
        name: str,
        label: str,
        current: int,
        fg: str = OneColors.BLACK,
        bg: str = OneColors.WHITE,
    ) -> None:
        self._states[name] = (label, current)

        def render():
            label_, current_ = self._states[name]
            text = f"{label_}: {current_}"
            return HTML(
                f"<style fg='{fg}' bg='{bg}'>{text:^{self.get_space()}}</style>"
            )

        self._add_named(name, render)

    def add_total_counter(
        self,
        name: str,
        label: str,
        current: int,
        total: int,
        fg: str = OneColors.BLACK,
        bg: str = OneColors.WHITE,
    ) -> None:
        self._states[name] = (label, current, total)

        if current > total:
            raise ValueError(
                f"Current value {current} is greater than total value {total}"
            )

        def render():
            label_, current_, text_ = self._states[name]
            text = f"{label_}: {current_}/{text_}"
            return HTML(
                f"<style fg='{fg}' bg='{bg}'>{text:^{self.get_space()}}</style>"
            )

        self._add_named(name, render)

    def add_toggle(
        self,
        key: str,
        label: str,
        state: bool,
        fg: str = OneColors.BLACK,
        bg_on: str = OneColors.GREEN,
        bg_off: str = OneColors.DARK_RED,
    ) -> None:
        key = key.upper()
        if key in self.toggles:
            raise ValueError(f"Key {key} is already used as a toggle")
        self._states[key] = (label, state)
        self.toggles.append(key)

        def render():
            label_, state_ = self._states[key]
            color = bg_on if state_ else bg_off
            status = "ON" if state_ else "OFF"
            text = f"({key.upper()}) {label_}: {status}"
            return HTML(
                f"<style bg='{color}' fg='{fg}'>{text:^{self.get_space()}}</style>"
            )

        self._add_named(key, render)

        for k in (key.upper(), key.lower()):

            @self.key_bindings.add(k)
            def _(event, key=k):
                self.toggle_state(key)

    def toggle_state(self, name: str) -> bool:
        label, state = self._states.get(name, (None, False))
        new_state = not state
        self.update_toggle(name, new_state)
        return new_state

    def update_toggle(self, name: str, state: bool) -> None:
        if name in self._states:
            label, _ = self._states[name]
            self._states[name] = (label, state)

    def increment_counter(self, name: str) -> None:
        if name in self._states:
            label, current = self._states[name]
            self._states[name] = (label, current + 1)

    def increment_total_counter(self, name: str) -> None:
        if name in self._states:
            label, current, total = self._states[name]
            if current < total:
                self._states[name] = (label, current + 1, total)

    def update_counter(
        self, name: str, current: Optional[int] = None, total: Optional[int] = None
    ) -> None:
        if name in self._states:
            label, c, t = self._states[name]
            self._states[name] = (
                label,
                current if current is not None else c,
                total if total is not None else t,
            )

    def _add_named(self, name: str, render_fn: Callable[[], HTML]) -> None:
        self._named_items[name] = render_fn
        self._items = list(self._named_items.values())

    def render(self):
        return merge_formatted_text([fn() for fn in self._items])



import asyncio
import logging
from rich.markdown import Markdown

from falyx import Action, Falyx, HookType
from falyx.hooks import log_before, log_success, log_error, log_after
from falyx.themes.colors import OneColors
from falyx.utils import setup_logging

# Setup logging
setup_logging(console_log_level=logging.WARNING, json_log_to_file=True)


def main():
    # Create the menu
    menu = Falyx(
        title=Markdown("# 🚀 Falyx CLI Demo"),
        welcome_message="Welcome to Falyx!",
        exit_message="Thanks for using Falyx!",
        include_history_command=True,
        include_help_command=True,
    )

    # Define async actions
    async def hello():
        print("👋 Hello from Falyx CLI!")

    def goodbye():
        print("👋 Goodbye from Falyx CLI!")

    async def do_task_and_increment(counter_name: str = "tasks"):
        await asyncio.sleep(3)
        print("✅ Task completed.")
        menu.bottom_bar.increment_total_counter(counter_name)

    # Register global logging hooks
    menu.hooks.register(HookType.BEFORE, log_before)
    menu.hooks.register(HookType.ON_SUCCESS, log_success)
    menu.hooks.register(HookType.ON_ERROR, log_error)
    menu.hooks.register(HookType.AFTER, log_after)

    # Add a toggle to the bottom bar
    menu.add_toggle("D", "Debug Mode", state=False)

    # Add a counter to the bottom bar
    menu.add_total_counter("tasks", "Tasks", current=0, total=5)

    # Add static text to the bottom bar
    menu.add_static("env", "🌐 Local Env")

    # Add commands with help_text
    menu.add_command(
        key="S",
        description="Say Hello",
        help_text="Greets the user with a friendly hello message.",
        action=Action("Hello", hello),
        color=OneColors.CYAN,
    )

    menu.add_command(
        key="G",
        description="Say Goodbye",
        help_text="Bids farewell and thanks the user for using the app.",
        action=Action("Goodbye", goodbye),
        color=OneColors.MAGENTA,
    )

    menu.add_command(
        key="T",
        description="Run a Task",
        aliases=["task", "run"],
        help_text="Performs a task and increments the counter by 1.",
        action=do_task_and_increment,
        args=("tasks",),
        color=OneColors.GREEN,
        spinner=True,
    )

    asyncio.run(menu.cli())


if __name__ == "__main__":
    """
    Entry point for the Falyx CLI demo application.
    This function initializes the menu and runs it.
    """
    main()


"""action.py

Any Action or Command is callable and supports the signature:
    result = thing(*args, **kwargs)

This guarantees:
- Hook lifecycle (before/after/error/teardown)
- Timing
- Consistent return values
"""
from __future__ import annotations

import asyncio
import random
from abc import ABC, abstractmethod
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from typing import Any, Callable

from rich.console import Console
from rich.tree import Tree

from falyx.context import ExecutionContext, ResultsContext
from falyx.debug import register_debug_hooks
from falyx.execution_registry import ExecutionRegistry as er
from falyx.hook_manager import Hook, HookManager, HookType
from falyx.retry import RetryHandler, RetryPolicy
from falyx.themes.colors import OneColors
from falyx.utils import ensure_async, logger

console = Console()


class BaseAction(ABC):
    """
    Base class for actions. Actions can be simple functions or more
    complex actions like `ChainedAction` or `ActionGroup`. They can also
    be run independently or as part of Menu.

    inject_last_result (bool): Whether to inject the previous action's result into kwargs.
    inject_last_result_as (str): The name of the kwarg key to inject the result as
                                 (default: 'last_result').
    """
    def __init__(
            self,
            name: str,
            hooks: HookManager | None = None,
            inject_last_result: bool = False,
            inject_last_result_as: str = "last_result",
            logging_hooks: bool = False,
    ) -> None:
        self.name = name
        self.hooks = hooks or HookManager()
        self.is_retryable: bool = False
        self.results_context: ResultsContext | None = None
        self.inject_last_result: bool = inject_last_result
        self.inject_last_result_as: str = inject_last_result_as

        if logging_hooks:
            register_debug_hooks(self.hooks)

    async def __call__(self, *args, **kwargs) -> Any:
        return await self._run(*args, **kwargs)

    @abstractmethod
    async def _run(self, *args, **kwargs) -> Any:
        raise NotImplementedError("_run must be implemented by subclasses")

    @abstractmethod
    async def preview(self, parent: Tree | None = None):
        raise NotImplementedError("preview must be implemented by subclasses")

    def set_results_context(self, results_context: ResultsContext):
        self.results_context = results_context

    def prepare_for_chain(self, results_context: ResultsContext) -> BaseAction:
        """
        Prepare the action specifically for sequential (ChainedAction) execution.
        Can be overridden for chain-specific logic.
        """
        self.set_results_context(results_context)
        return self

    def prepare_for_group(self, results_context: ResultsContext) -> BaseAction:
        """
        Prepare the action specifically for parallel (ActionGroup) execution.
        Can be overridden for group-specific logic.
        """
        self.set_results_context(results_context)
        return self

    def _maybe_inject_last_result(self, kwargs: dict[str, Any]) -> dict[str, Any]:
        if self.inject_last_result and self.results_context:
            key = self.inject_last_result_as
            if key in kwargs:
                logger.warning("[%s] ⚠️ Overriding '%s' with last_result", self.name, key)
            kwargs = dict(kwargs)
            kwargs[key] = self.results_context.last_result()
        return kwargs

    def register_hooks_recursively(self, hook_type: HookType, hook: Hook):
        """Register a hook for all actions and sub-actions."""
        self.hooks.register(hook_type, hook)

    def __str__(self):
        return f"<{self.__class__.__name__} '{self.name}'>"

    def __repr__(self):
        return str(self)

    @classmethod
    def enable_retries_recursively(cls, action: BaseAction, policy: RetryPolicy | None):
        if not policy:
            policy = RetryPolicy(enabled=True)
        if isinstance(action, Action):
            action.retry_policy = policy
            action.retry_policy.enabled = True
            action.hooks.register("on_error", RetryHandler(policy).retry_on_error)

        if hasattr(action, "actions"):
            for sub in action.actions:
                cls.enable_retries_recursively(sub, policy)


class Action(BaseAction):
    """A simple action that runs a callable. It can be a function or a coroutine."""
    def __init__(
            self,
            name: str,
            action,
            rollback=None,
            args: tuple[Any, ...] = (),
            kwargs: dict[str, Any] | None = None,
            hooks: HookManager | None = None,
            inject_last_result: bool = False,
            inject_last_result_as: str = "last_result",
            retry: bool = False,
            retry_policy: RetryPolicy | None = None,
    ) -> None:
        super().__init__(name, hooks, inject_last_result, inject_last_result_as)
        self.action = ensure_async(action)
        self.rollback = rollback
        self.args = args
        self.kwargs = kwargs or {}
        self.is_retryable = True
        self.retry_policy = retry_policy or RetryPolicy()
        if retry or (retry_policy and retry_policy.enabled):
            self.enable_retry()

    def enable_retry(self):
        """Enable retry with the existing retry policy."""
        self.retry_policy.enabled = True
        logger.debug(f"[Action:{self.name}] Registering retry handler")
        handler = RetryHandler(self.retry_policy)
        self.hooks.register(HookType.ON_ERROR, handler.retry_on_error)

    def set_retry_policy(self, policy: RetryPolicy):
        """Set a new retry policy and re-register the handler."""
        self.retry_policy = policy
        self.enable_retry()

    async def _run(self, *args, **kwargs) -> Any:
        combined_args = args + self.args
        combined_kwargs = self._maybe_inject_last_result({**self.kwargs, **kwargs})

        context = ExecutionContext(
            name=self.name,
            args=combined_args,
            kwargs=combined_kwargs,
            action=self,
        )
        context.start_timer()
        try:
            await self.hooks.trigger(HookType.BEFORE, context)
            result = await self.action(*combined_args, **combined_kwargs)
            context.result = result
            await self.hooks.trigger(HookType.ON_SUCCESS, context)
            return context.result
        except Exception as error:
            context.exception = error
            await self.hooks.trigger(HookType.ON_ERROR, context)
            if context.result is not None:
                logger.info("[%s] ✅ Recovered: %s", self.name, self.name)
                return context.result
            raise error
        finally:
            context.stop_timer()
            await self.hooks.trigger(HookType.AFTER, context)
            await self.hooks.trigger(HookType.ON_TEARDOWN, context)
            er.record(context)

    async def preview(self, parent: Tree | None = None):
        label = [f"[{OneColors.GREEN_b}]⚙ Action[/] '{self.name}'"]
        if self.inject_last_result:
            label.append(f" [dim](injects '{self.inject_last_result_as}')[/dim]")
        if self.retry_policy.enabled:
            label.append(
                f"\n[dim]↻ Retries:[/] {self.retry_policy.max_retries}x, "
                f"delay {self.retry_policy.delay}s, backoff {self.retry_policy.backoff}x"
            )

        if parent:
            parent.add("".join(label))
        else:
            console.print(Tree("".join(label)))


class ActionListMixin:
    """Mixin for managing a list of actions."""
    def __init__(self) -> None:
        self.actions: list[BaseAction] = []

    def set_actions(self, actions: list[BaseAction]) -> None:
        """Replaces the current action list with a new one."""
        self.actions.clear()
        for action in actions:
            self.add_action(action)

    def add_action(self, action: BaseAction) -> None:
        """Adds an action to the list."""
        self.actions.append(action)

    def remove_action(self, name: str) -> None:
        """Removes an action by name."""
        self.actions = [action for action in self.actions if action.name != name]

    def has_action(self, name: str) -> bool:
        """Checks if an action with the given name exists."""
        return any(action.name == name for action in self.actions)

    def get_action(self, name: str) -> BaseAction | None:
        """Retrieves an action by name."""
        for action in self.actions:
            if action.name == name:
                return action
        return None


class ChainedAction(BaseAction, ActionListMixin):
    """A ChainedAction is a sequence of actions that are executed in order."""
    def __init__(
            self,
            name: str,
            actions: list[BaseAction] | None = None,
            hooks: HookManager | None = None,
            inject_last_result: bool = False,
            inject_last_result_as: str = "last_result",
    ) -> None:
        super().__init__(name, hooks, inject_last_result, inject_last_result_as)
        ActionListMixin.__init__(self)
        if actions:
            self.set_actions(actions)

    async def _run(self, *args, **kwargs) -> list[Any]:
        results_context = ResultsContext(name=self.name)
        if self.results_context:
            results_context.add_result(self.results_context.last_result())
        updated_kwargs = self._maybe_inject_last_result(kwargs)
        context = ExecutionContext(
            name=self.name,
            args=args,
            kwargs=updated_kwargs,
            action=self,
            extra={"results": [], "rollback_stack": []},
        )
        context.start_timer()
        try:
            await self.hooks.trigger(HookType.BEFORE, context)

            for index, action in enumerate(self.actions):
                results_context.current_index = index
                prepared = action.prepare_for_chain(results_context)
                result = await prepared(*args, **updated_kwargs)
                results_context.add_result(result)
                context.extra["results"].append(result)
                context.extra["rollback_stack"].append(prepared)

            context.result = context.extra["results"]
            await self.hooks.trigger(HookType.ON_SUCCESS, context)
            return context.result

        except Exception as error:
            context.exception = error
            results_context.errors.append((results_context.current_index, error))
            await self._rollback(context.extra["rollback_stack"], *args, **kwargs)
            await self.hooks.trigger(HookType.ON_ERROR, context)
            raise
        finally:
            context.stop_timer()
            await self.hooks.trigger(HookType.AFTER, context)
            await self.hooks.trigger(HookType.ON_TEARDOWN, context)
            er.record(context)

    async def _rollback(self, rollback_stack, *args, **kwargs):
        for action in reversed(rollback_stack):
            rollback = getattr(action, "rollback", None)
            if rollback:
                try:
                    logger.warning("[%s] ↩️ Rolling back...", action.name)
                    await action.rollback(*args, **kwargs)
                except Exception as error:
                    logger.error("[%s]⚠️ Rollback failed: %s", action.name, error)

    async def preview(self, parent: Tree | None = None):
        label = f"[{OneColors.CYAN_b}]⛓ ChainedAction[/] '{self.name}'"
        if self.inject_last_result:
            label.append(f" [dim](injects '{self.inject_last_result_as}')[/dim]")
        tree = parent.add("".join(label)) if parent else Tree("".join(label))
        for action in self.actions:
            await action.preview(parent=tree)
        if not parent:
            console.print(tree)

    def register_hooks_recursively(self, hook_type: HookType, hook: Hook):
        """Register a hook for all actions and sub-actions."""
        self.hooks.register(hook_type, hook)
        for action in self.actions:
            action.register_hooks_recursively(hook_type, hook)


class ActionGroup(BaseAction, ActionListMixin):
    """An ActionGroup is a collection of actions that can be run in parallel."""
    def __init__(
            self,
            name: str,
            actions: list[BaseAction] | None = None,
            hooks: HookManager | None = None,
            inject_last_result: bool = False,
            inject_last_result_as: str = "last_result",
    ):
        super().__init__(name, hooks, inject_last_result, inject_last_result_as)
        ActionListMixin.__init__(self)
        if actions:
            self.set_actions(actions)

    async def _run(self, *args, **kwargs) -> list[tuple[str, Any]]:
        results_context = ResultsContext(name=self.name, is_parallel=True)
        if self.results_context:
            results_context.set_shared_result(self.results_context.last_result())
        updated_kwargs = self._maybe_inject_last_result(kwargs)
        context = ExecutionContext(
            name=self.name,
            args=args,
            kwargs=updated_kwargs,
            action=self,
            extra={"results": [], "errors": []},
        )
        async def run_one(action: BaseAction):
            try:
                prepared = action.prepare_for_group(results_context)
                result = await prepared(*args, **updated_kwargs)
                results_context.add_result((action.name, result))
                context.extra["results"].append((action.name, result))
            except Exception as error:
                results_context.errors.append((results_context.current_index, error))
                context.extra["errors"].append((action.name, error))

        context.start_timer()
        try:
            await self.hooks.trigger(HookType.BEFORE, context)
            await asyncio.gather(*[run_one(a) for a in self.actions])

            if context.extra["errors"]:
                context.exception = Exception(
                    f"{len(context.extra['errors'])} action(s) failed: "
                    f"{' ,'.join(name for name, _ in context.extra["errors"])}"
                )
                await self.hooks.trigger(HookType.ON_ERROR, context)
                raise context.exception

            context.result = context.extra["results"]
            await self.hooks.trigger(HookType.ON_SUCCESS, context)
            return context.result

        except Exception as error:
            context.exception = error
            raise
        finally:
            context.stop_timer()
            await self.hooks.trigger(HookType.AFTER, context)
            await self.hooks.trigger(HookType.ON_TEARDOWN, context)
            er.record(context)

    async def preview(self, parent: Tree | None = None):
        label = [f"[{OneColors.MAGENTA_b}]⏩ ActionGroup (parallel)[/] '{self.name}'"]
        if self.inject_last_result:
            label.append(f" [dim](receives '{self.inject_last_result_as}')[/dim]")
        tree = parent.add("".join(label)) if parent else Tree("".join(label))
        actions = self.actions.copy()
        random.shuffle(actions)
        await asyncio.gather(*(action.preview(parent=tree) for action in actions))
        if not parent:
            console.print(tree)

    def register_hooks_recursively(self, hook_type: HookType, hook: Hook):
        """Register a hook for all actions and sub-actions."""
        super().register_hooks_recursively(hook_type, hook)
        for action in self.actions:
            action.register_hooks_recursively(hook_type, hook)


class ProcessAction(BaseAction):
    """A ProcessAction runs a function in a separate process using ProcessPoolExecutor."""
    def __init__(
        self,
        name: str,
        func: Callable[..., Any],
        args: tuple = (),
        kwargs: dict[str, Any] | None = None,
        hooks: HookManager | None = None,
        executor: ProcessPoolExecutor | None = None,
        inject_last_result: bool = False,
        inject_last_result_as: str = "last_result",
    ):
        super().__init__(name, hooks, inject_last_result, inject_last_result_as)
        self.func = func
        self.args = args
        self.kwargs = kwargs or {}
        self.executor = executor or ProcessPoolExecutor()
        self.is_retryable = True

    async def _run(self, *args, **kwargs):
        if self.inject_last_result:
            last_result = self.results_context.last_result()
            if not self._validate_pickleable(last_result):
                raise ValueError(
                    f"Cannot inject last result into {self.name}: "
                    f"last result is not pickleable."
                )
        combined_args = args + self.args
        combined_kwargs = self._maybe_inject_last_result({**self.kwargs, **kwargs})
        context = ExecutionContext(
            name=self.name,
            args=combined_args,
            kwargs=combined_kwargs,
            action=self,
        )
        loop = asyncio.get_running_loop()

        context.start_timer()
        try:
            await self.hooks.trigger(HookType.BEFORE, context)
            result = await loop.run_in_executor(
                self.executor, partial(self.func, *combined_args, **combined_kwargs)
            )
            context.result = result
            await self.hooks.trigger(HookType.ON_SUCCESS, context)
            return result
        except Exception as error:
            context.exception = error
            await self.hooks.trigger(HookType.ON_ERROR, context)
            if context.result is not None:
                return context.result
            raise
        finally:
            context.stop_timer()
            await self.hooks.trigger(HookType.AFTER, context)
            await self.hooks.trigger(HookType.ON_TEARDOWN, context)
            er.record(context)

    async def preview(self, parent: Tree | None = None):
        label = [f"[{OneColors.DARK_YELLOW_b}]🧠 ProcessAction (new process)[/] '{self.name}'"]
        if self.inject_last_result:
            label.append(f" [dim](injects '{self.inject_last_result_as}')[/dim]")
        if parent:
            parent.add("".join(label))
        else:
            console.print(Tree("".join(label)))

    def _validate_pickleable(self, obj: Any) -> bool:
        try:
            import pickle
            pickle.dumps(obj)
            print("YES")
            return True
        except (pickle.PicklingError, TypeError):
            print("NO")
            return False



"""parsers.py
This module contains the argument parsers used for the Falyx CLI.
"""
from dataclasses import asdict, dataclass
from argparse import ArgumentParser


@dataclass
class FalyxParsers:
    """Defines the argument parsers for the Falyx CLI."""
    root: ArgumentParser
    run: ArgumentParser
    run_all: ArgumentParser
    preview: ArgumentParser
    list: ArgumentParser
    version: ArgumentParser

    def as_dict(self) -> dict[str, ArgumentParser]:
        """Convert the FalyxParsers instance to a dictionary."""
        return asdict(self)

    def get_parser(self, name: str) -> ArgumentParser | None:
        """Get the parser by name."""
        return self.as_dict().get(name)


def get_arg_parsers() -> FalyxParsers:
    """Returns the argument parser for the CLI."""
    parser = ArgumentParser(prog="falyx", description="Falyx CLI - Run structured async command workflows.")
    parser.add_argument("-v", "--verbose", action="store_true", help="Enable debug logging for Falyx.")
    parser.add_argument("--debug-hooks", action="store_true", help="Enable default lifecycle debug logging")
    parser.add_argument("--version", action="store_true", help="Show Falyx version")
    subparsers = parser.add_subparsers(dest="command")

    run_parser = subparsers.add_parser("run", help="Run a specific command")
    run_parser.add_argument("name", help="Key, alias, or description of the command")
    run_parser.add_argument("--retries", type=int, help="Number of retries on failure", default=0)
    run_parser.add_argument("--retry-delay", type=float, help="Initial delay between retries in (seconds)", default=0)
    run_parser.add_argument("--retry-backoff", type=float, help="Backoff factor for retries", default=0)
    run_group = run_parser.add_mutually_exclusive_group(required=False)
    run_group.add_argument("-c", "--confirm", dest="force_confirm", action="store_true", help="Force confirmation prompts")
    run_group.add_argument("-s", "--skip-confirm", dest="skip_confirm", action="store_true", help="Skip confirmation prompts")

    run_all_parser = subparsers.add_parser("run-all", help="Run all commands with a given tag")
    run_all_parser.add_argument("-t", "--tag", required=True, help="Tag to match")
    run_all_parser.add_argument("--retries", type=int, help="Number of retries on failure", default=0)
    run_all_parser.add_argument("--retry-delay", type=float, help="Initial delay between retries in (seconds)", default=0)
    run_all_parser.add_argument("--retry-backoff", type=float, help="Backoff factor for retries", default=0)
    run_all_group = run_all_parser.add_mutually_exclusive_group(required=False)
    run_all_group.add_argument("-c", "--confirm", dest="force_confirm", action="store_true", help="Force confirmation prompts")
    run_all_group.add_argument("-s", "--skip-confirm", dest="skip_confirm", action="store_true", help="Skip confirmation prompts")

    preview_parser = subparsers.add_parser("preview", help="Preview a command without running it")
    preview_parser.add_argument("name", help="Key, alias, or description of the command")

    list_parser = subparsers.add_parser("list", help="List all available commands with tags")

    version_parser = subparsers.add_parser("version", help="Show the Falyx version")

    return FalyxParsers(
        root=parser,
        run=run_parser,
        run_all=run_all_parser,
        preview=preview_parser,
        list=list_parser,
        version=version_parser,
    )



"""falyx.py

This class creates a Falyx object that creates a selectable menu
with customizable commands and functionality.

It allows for adding commands, and their accompanying actions,
and provides a method to display the menu and handle user input.

This class uses the `rich` library to display the menu in a
formatted and visually appealing way.

This class also uses the `prompt_toolkit` library to handle
user input and create an interactive experience.
"""
import asyncio
import logging
import sys
from argparse import Namespace
from difflib import get_close_matches
from functools import cached_property
from typing import Any, Callable

from prompt_toolkit import PromptSession
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.formatted_text import AnyFormattedText
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.validation import Validator
from rich import box
from rich.console import Console
from rich.markdown import Markdown
from rich.table import Table

from falyx.action import BaseAction
from falyx.bottom_bar import BottomBar
from falyx.command import Command
from falyx.context import ExecutionContext
from falyx.debug import log_after, log_before, log_error, log_success
from falyx.exceptions import (CommandAlreadyExistsError, FalyxError,
                              InvalidActionError, NotAFalyxError)
from falyx.execution_registry import ExecutionRegistry as er
from falyx.hook_manager import Hook, HookManager, HookType
from falyx.parsers import FalyxParsers, get_arg_parsers
from falyx.retry import RetryPolicy
from falyx.themes.colors import OneColors, get_nord_theme
from falyx.utils import CaseInsensitiveDict, async_confirm, chunks, logger
from falyx.version import __version__


class Falyx:
    """Class to create a menu with commands.

    Hook functions must have the signature:
        def hook(command: Command) -> None:
    where `command` is the selected command.

    Error hook functions must have the signature:
        def error_hook(command: Command, error: Exception) -> None:
    where `command` is the selected command and `error` is the exception raised.

    Hook execution order:
    1. Before action hooks of the menu.
    2. Before action hooks of the selected command.
    3. Action of the selected command.
    4. After action hooks of the selected command.
    5. After action hooks of the menu.
    6. On error hooks of the selected command (if an error occurs).
    7. On error hooks of the menu (if an error occurs).

    Parameters:
        title (str|Markdown): The title of the menu.
        columns (int): The number of columns to display the commands in.
        prompt (AnyFormattedText): The prompt to display when asking for input.
        bottom_bar (str|callable|None): The text to display in the bottom bar.
    """

    def __init__(
        self,
        title: str | Markdown = "Menu",
        prompt: str | AnyFormattedText = "> ",
        columns: int = 3,
        bottom_bar: BottomBar | str | Callable[[], None] | None = None,
        welcome_message: str | Markdown | dict[str, Any] = "",
        exit_message: str | Markdown | dict[str, Any] = "",
        key_bindings: KeyBindings | None = None,
        include_history_command: bool = True,
        include_help_command: bool = False,
        confirm_on_error: bool = True,
        never_confirm: bool = False,
        always_confirm: bool = False,
        cli_args: Namespace | None = None,
        custom_table: Callable[["Falyx"], Table] | Table | None = None,
    ) -> None:
        """Initializes the Falyx object."""
        self.title: str | Markdown = title
        self.prompt: str | AnyFormattedText = prompt
        self.columns: int = columns
        self.commands: dict[str, Command] = CaseInsensitiveDict()
        self.exit_command: Command = self._get_exit_command()
        self.history_command: Command | None = self._get_history_command() if include_history_command else None
        self.help_command: Command | None = self._get_help_command() if include_help_command else None
        self.console: Console = Console(color_system="truecolor", theme=get_nord_theme())
        self.welcome_message: str | Markdown | dict[str, Any] = welcome_message
        self.exit_message: str | Markdown | dict[str, Any] = exit_message
        self.hooks: HookManager = HookManager()
        self.last_run_command: Command | None = None
        self.key_bindings: KeyBindings = key_bindings or KeyBindings()
        self.bottom_bar: BottomBar | str | Callable[[], None] = bottom_bar or BottomBar(columns=columns, key_bindings=self.key_bindings)
        self.confirm_on_error: bool = confirm_on_error
        self._never_confirm: bool = never_confirm
        self._always_confirm: bool = always_confirm
        self.cli_args: Namespace | None = cli_args
        self.custom_table: Callable[["Falyx"], Table] | Table | None = custom_table

    @property
    def _name_map(self) -> dict[str, Command]:
        """Builds a mapping of all valid input names (keys, aliases, normalized names) to Command objects.
        If a collision occurs, logs a warning and keeps the first registered command.
        """
        mapping: dict[str, Command] = {}

        def register(name: str, cmd: Command):
            norm = name.upper().strip()
            if norm in mapping:
                existing = mapping[norm]
                if existing is not cmd:
                    logger.warning(
                        f"[alias conflict] '{name}' already assigned to '{existing.description}'."
                        f" Skipping for '{cmd.description}'."
                    )
            else:
                mapping[norm] = cmd

        for special in [self.exit_command, self.history_command, self.help_command]:
            if special:
                register(special.key, special)
                for alias in special.aliases:
                    register(alias, special)
                register(special.description, special)

        for cmd in self.commands.values():
            register(cmd.key, cmd)
            for alias in cmd.aliases:
                register(alias, cmd)
            register(cmd.description, cmd)
        return mapping

    def get_title(self) -> str:
        """Returns the string title of the menu."""
        if isinstance(self.title, str):
            return self.title
        elif isinstance(self.title, Markdown):
            return self.title.markup
        return self.title

    def _get_exit_command(self) -> Command:
        """Returns the back command for the menu."""
        return Command(
            key="Q",
            description="Exit",
            aliases=["EXIT", "QUIT"],
            color=OneColors.DARK_RED,
        )

    def _get_history_command(self) -> Command:
        """Returns the history command for the menu."""
        return Command(
            key="Y",
            description="History",
            aliases=["HISTORY"],
            action=er.get_history_action(),
            color=OneColors.DARK_YELLOW,
        )

    async def _show_help(self):
        table = Table(title="[bold cyan]Help Menu[/]", box=box.SIMPLE)
        table.add_column("Key", style="bold", no_wrap=True)
        table.add_column("Aliases", style="dim", no_wrap=True)
        table.add_column("Description", style="dim", overflow="fold")
        table.add_column("Tags", style="dim", no_wrap=True)

        for command in self.commands.values():
            help_text = command.help_text or command.description
            table.add_row(
                f"[{command.color}]{command.key}[/]",
                ", ".join(command.aliases) if command.aliases else "None",
                help_text,
                ", ".join(command.tags) if command.tags else "None"
            )

        table.add_row(
            f"[{self.exit_command.color}]{self.exit_command.key}[/]",
            ", ".join(self.exit_command.aliases),
            "Exit this menu or program"
        )

        if self.history_command:
            table.add_row(
                f"[{self.history_command.color}]{self.history_command.key}[/]",
                ", ".join(self.history_command.aliases),
                "History of executed actions"
        )

        if self.help_command:
            table.add_row(
                f"[{self.help_command.color}]{self.help_command.key}[/]",
                ", ".join(self.help_command.aliases),
                "Show this help menu"
            )

        self.console.print(table)

    def _get_help_command(self) -> Command:
        """Returns the help command for the menu."""
        return Command(
            key="H",
            aliases=["HELP"],
            description="Help",
            action=self._show_help,
            color=OneColors.LIGHT_YELLOW,
        )
    def _get_completer(self) -> WordCompleter:
        """Completer to provide auto-completion for the menu commands."""
        keys = [self.exit_command.key]
        keys.extend(self.exit_command.aliases)
        if self.history_command:
            keys.append(self.history_command.key)
            keys.extend(self.history_command.aliases)
        if self.help_command:
            keys.append(self.help_command.key)
            keys.extend(self.help_command.aliases)
        for cmd in self.commands.values():
            keys.append(cmd.key)
            keys.extend(cmd.aliases)
        return WordCompleter(keys, ignore_case=True)

    def _get_validator(self) -> Validator:
        """Validator to check if the input is a valid command or toggle key."""
        keys = {self.exit_command.key.upper()}
        keys.update({alias.upper() for alias in self.exit_command.aliases})
        if self.history_command:
            keys.add(self.history_command.key.upper())
            keys.update({alias.upper() for alias in self.history_command.aliases})
        if self.help_command:
            keys.add(self.help_command.key.upper())
            keys.update({alias.upper() for alias in self.help_command.aliases})

        for cmd in self.commands.values():
            keys.add(cmd.key.upper())
            keys.update({alias.upper() for alias in cmd.aliases})

        if isinstance(self.bottom_bar, BottomBar):
            toggle_keys = {key.upper() for key in self.bottom_bar.toggles}
        else:
            toggle_keys = set()

        commands_str = ", ".join(sorted(keys))
        toggles_str = ", ".join(sorted(toggle_keys))

        message_lines = ["Invalid input. Available keys:"]
        if keys:
            message_lines.append(f"  Commands: {commands_str}")
        if toggle_keys:
            message_lines.append(f"  Toggles: {toggles_str}")
        error_message = " ".join(message_lines)

        def validator(text):
            return True if self.get_command(text, from_validate=True) else False

        return Validator.from_callable(
            validator,
            error_message=error_message,
            move_cursor_to_end=True,
        )

    def _invalidate_session_cache(self):
        """Forces the session to be recreated on the next access."""
        if hasattr(self, "session"):
            del self.session

    def add_toggle(self, key: str, label: str, state: bool) -> None:
        """Adds a toggle to the bottom bar."""
        assert isinstance(self.bottom_bar, BottomBar), "Bottom bar must be an instance of BottomBar."
        self.bottom_bar.add_toggle(key, label, state)
        self._invalidate_session_cache()

    def add_counter(self, name: str, label: str, current: int) -> None:
        """Adds a counter to the bottom bar."""
        assert isinstance(self.bottom_bar, BottomBar), "Bottom bar must be an instance of BottomBar."
        self.bottom_bar.add_counter(name, label, current)
        self._invalidate_session_cache()

    def add_total_counter(self, name: str, label: str, current: int, total: int) -> None:
        """Adds a counter to the bottom bar."""
        assert isinstance(self.bottom_bar, BottomBar), "Bottom bar must be an instance of BottomBar."
        self.bottom_bar.add_total_counter(name, label, current, total)
        self._invalidate_session_cache()

    def add_static(self, name: str, text: str) -> None:
        """Adds a static element to the bottom bar."""
        assert isinstance(self.bottom_bar, BottomBar), "Bottom bar must be an instance of BottomBar."
        self.bottom_bar.add_static(name, text)
        self._invalidate_session_cache

    def get_toggle_state(self, key: str) -> bool | None:
        assert isinstance(self.bottom_bar, BottomBar), "Bottom bar must be an instance of BottomBar."
        if key.upper() in self.bottom_bar._states:
            """Returns the state of a toggle."""
            return self.bottom_bar._states[key.upper()][1]
        return None

    def add_help_command(self):
        """Adds a help command to the menu if it doesn't already exist."""
        if not self.help_command:
            self.help_command = self._get_help_command()
            self._invalidate_session_cache()

    def add_history_command(self):
        """Adds a history command to the menu if it doesn't already exist."""
        if not self.history_command:
            self.history_command = self._get_history_command()
            self._invalidate_session_cache()

    def _get_bottom_bar(self) -> Callable[[], Any] | str | None:
        """Returns the bottom bar for the menu."""
        if isinstance(self.bottom_bar, BottomBar) and self.bottom_bar._items:
            return self.bottom_bar.render
        elif callable(self.bottom_bar):
            return self.bottom_bar
        elif isinstance(self.bottom_bar, str):
            return self.bottom_bar
        return None

    @cached_property
    def session(self) -> PromptSession:
        """Returns the prompt session for the menu."""
        return PromptSession(
            message=self.prompt,
            multiline=False,
            completer=self._get_completer(),
            reserve_space_for_menu=1,
            validator=self._get_validator(),
            bottom_toolbar=self._get_bottom_bar(),
            key_bindings=self.key_bindings,
        )

    def register_all_hooks(self, hook_type: HookType, hooks: Hook | list[Hook]) -> None:
        """Registers hooks for all commands in the menu and actions recursively."""
        hook_list = hooks if isinstance(hooks, list) else [hooks]
        for hook in hook_list:
            if not callable(hook):
                raise InvalidActionError("Hook must be a callable.")
            self.hooks.register(hook_type, hook)
            for command in self.commands.values():
                command.hooks.register(hook_type, hook)
                if isinstance(command.action, Falyx):
                    command.action.register_all_hooks(hook_type, hook)
                if isinstance(command.action, BaseAction):
                    command.action.register_hooks_recursively(hook_type, hook)

    def register_all_with_debug_hooks(self) -> None:
        """Registers debug hooks for all commands in the menu and actions recursively."""
        self.register_all_hooks(HookType.BEFORE, log_before)
        self.register_all_hooks(HookType.ON_SUCCESS, log_success)
        self.register_all_hooks(HookType.ON_ERROR, log_error)
        self.register_all_hooks(HookType.AFTER, log_after)

    def debug_hooks(self) -> None:
        """Logs the names of all hooks registered for the menu and its commands."""
        def hook_names(hook_list):
            return [hook.__name__ for hook in hook_list]

        logger.debug(f"Menu-level before hooks: {hook_names(self.hooks._hooks[HookType.BEFORE])}")
        logger.debug(f"Menu-level success hooks: {hook_names(self.hooks._hooks[HookType.ON_SUCCESS])}")
        logger.debug(f"Menu-level error hooks: {hook_names(self.hooks._hooks[HookType.ON_ERROR])}")
        logger.debug(f"Menu-level after hooks: {hook_names(self.hooks._hooks[HookType.AFTER])}")
        logger.debug(f"Menu-level on_teardown hooks: {hook_names(self.hooks._hooks[HookType.ON_TEARDOWN])}")

        for key, command in self.commands.items():
            logger.debug(f"[Command '{key}'] before: {hook_names(command.hooks._hooks[HookType.BEFORE])}")
            logger.debug(f"[Command '{key}'] success: {hook_names(command.hooks._hooks[HookType.ON_SUCCESS])}")
            logger.debug(f"[Command '{key}'] error: {hook_names(command.hooks._hooks[HookType.ON_ERROR])}")
            logger.debug(f"[Command '{key}'] after: {hook_names(command.hooks._hooks[HookType.AFTER])}")
            logger.debug(f"[Command '{key}'] on_teardown: {hook_names(command.hooks._hooks[HookType.ON_TEARDOWN])}")

    def _validate_command_key(self, key: str) -> None:
        """Validates the command key to ensure it is unique."""
        key = key.upper()
        toggles = self.bottom_bar.toggles if isinstance(self.bottom_bar, BottomBar) else []
        collisions = []

        if key in self.commands:
            collisions.append("command")
        if key == self.exit_command.key.upper():
            collisions.append("back command")
        if self.history_command and key == self.history_command.key.upper():
            collisions.append("history command")
        if self.help_command and key == self.help_command.key.upper():
            collisions.append("help command")
        if key in toggles:
            collisions.append("toggle")

        if collisions:
            raise CommandAlreadyExistsError(f"Command key '{key}' conflicts with existing {', '.join(collisions)}.")

    def update_exit_command(
        self,
        key: str = "0",
        description: str = "Exit",
        action: Callable[[], Any] = lambda: None,
        color: str = OneColors.DARK_RED,
        confirm: bool = False,
        confirm_message: str = "Are you sure?",
    ) -> None:
        """Updates the back command of the menu."""
        if not callable(action):
            raise InvalidActionError("Action must be a callable.")
        self._validate_command_key(key)
        self.exit_command = Command(
            key=key,
            description=description,
            action=action,
            color=color,
            confirm=confirm,
            confirm_message=confirm_message,
        )
        self._invalidate_session_cache()

    def add_submenu(self, key: str, description: str, submenu: "Falyx", color: str = OneColors.CYAN) -> None:
        """Adds a submenu to the menu."""
        if not isinstance(submenu, Falyx):
            raise NotAFalyxError("submenu must be an instance of Falyx.")
        self._validate_command_key(key)
        self.add_command(key, description, submenu.menu, color=color)
        self._invalidate_session_cache()

    def add_commands(self, commands: list[dict]) -> None:
        """Adds multiple commands to the menu."""
        for command in commands:
            self.add_command(**command)

    def add_command(
        self,
        key: str,
        description: str,
        action: BaseAction | Callable[[], Any],
        aliases: list[str] | None = None,
        args: tuple = (),
        kwargs: dict[str, Any] = {},
        help_text: str = "",
        color: str = OneColors.WHITE,
        confirm: bool = False,
        confirm_message: str = "Are you sure?",
        preview_before_confirm: bool = True,
        spinner: bool = False,
        spinner_message: str = "Processing...",
        spinner_type: str = "dots",
        spinner_style: str = OneColors.CYAN,
        spinner_kwargs: dict[str, Any] | None = None,
        hooks: HookManager | None = None,
        before_hooks: list[Callable] | None = None,
        success_hooks: list[Callable] | None = None,
        after_hooks: list[Callable] | None = None,
        error_hooks: list[Callable] | None = None,
        teardown_hooks: list[Callable] | None = None,
        tags: list[str] | None = None,
        logging_hooks: bool = False,
        retry: bool = False,
        retry_all: bool = False,
        retry_policy: RetryPolicy | None = None,
    ) -> Command:
        """Adds an command to the menu, preventing duplicates."""
        self._validate_command_key(key)
        command = Command(
            key=key,
            description=description,
            aliases=aliases if aliases else [],
            help_text=help_text,
            action=action,
            args=args,
            kwargs=kwargs,
            color=color,
            confirm=confirm,
            confirm_message=confirm_message,
            preview_before_confirm=preview_before_confirm,
            spinner=spinner,
            spinner_message=spinner_message,
            spinner_type=spinner_type,
            spinner_style=spinner_style,
            spinner_kwargs=spinner_kwargs or {},
            tags=tags if tags else [],
            logging_hooks=logging_hooks,
            retry=retry,
            retry_all=retry_all,
            retry_policy=retry_policy or RetryPolicy(),
        )

        if hooks:
            if not isinstance(hooks, HookManager):
                raise NotAFalyxError("hooks must be an instance of HookManager.")
            command.hooks = hooks

        for hook in before_hooks or []:
            command.hooks.register(HookType.BEFORE, hook)
        for hook in success_hooks or []:
            command.hooks.register(HookType.ON_SUCCESS, hook)
        for hook in error_hooks or []:
            command.hooks.register(HookType.ON_ERROR, hook)
        for hook in after_hooks or []:
            command.hooks.register(HookType.AFTER, hook)
        for hook in teardown_hooks or []:
            command.hooks.register(HookType.ON_TEARDOWN, hook)

        self.commands[key] = command
        self._invalidate_session_cache()
        return command

    def get_bottom_row(self) -> list[str]:
        """Returns the bottom row of the table for displaying additional commands."""
        bottom_row = []
        if self.history_command:
            bottom_row.append(f"[{self.history_command.key}] [{self.history_command.color}]{self.history_command.description}")
        if self.help_command:
            bottom_row.append(f"[{self.help_command.key}] [{self.help_command.color}]{self.help_command.description}")
        bottom_row.append(f"[{self.exit_command.key}] [{self.exit_command.color}]{self.exit_command.description}")
        return bottom_row

    def build_default_table(self) -> Table:
        """Build the standard table layout. Developers can subclass or call this in custom tables."""
        table = Table(title=self.title, show_header=False, box=box.SIMPLE, expand=True)
        for chunk in chunks(self.commands.items(), self.columns):
            row = []
            for key, command in chunk:
                row.append(f"[{key}] [{command.color}]{command.description}")
            table.add_row(*row)
        bottom_row = self.get_bottom_row()
        table.add_row(*bottom_row)
        return table

    @property
    def table(self) -> Table:
        """Creates or returns a custom table to display the menu commands."""
        if callable(self.custom_table):
            return self.custom_table(self)
        elif isinstance(self.custom_table, Table):
            return self.custom_table
        else:
            return self.build_default_table()

    def get_command(self, choice: str, from_validate=False) -> Command | None:
        """Returns the selected command based on user input. Supports keys, aliases, and abbreviations."""
        choice = choice.upper()
        name_map = self._name_map

        if choice in name_map:
            return name_map[choice]

        prefix_matches = [cmd for key, cmd in name_map.items() if key.startswith(choice)]
        if len(prefix_matches) == 1:
            return prefix_matches[0]

        fuzzy_matches = get_close_matches(choice, list(name_map.keys()), n=3, cutoff=0.7)
        if fuzzy_matches:
            if not from_validate:
                self.console.print(f"[{OneColors.LIGHT_YELLOW}]⚠️ Unknown command '{choice}'. Did you mean:[/] ")
            for match in fuzzy_matches:
                cmd = name_map[match]
                self.console.print(f"  • [bold]{match}[/] → {cmd.description}")
        else:
            if not from_validate:
                self.console.print(f"[{OneColors.LIGHT_YELLOW}]⚠️ Unknown command '{choice}'[/]")
        return None

    async def _should_run_action(self, selected_command: Command) -> bool:
        if self._never_confirm:
            return True

        if self.cli_args and getattr(self.cli_args, "skip_confirm", False):
            return True

        if (self._always_confirm or
            selected_command.confirm or
            self.cli_args and getattr(self.cli_args, "force_confirm", False)
        ):
            if selected_command.preview_before_confirm:
                await selected_command.preview()
            confirm_answer = await async_confirm(selected_command.confirmation_prompt)

            if confirm_answer:
                logger.info(f"[{OneColors.LIGHT_YELLOW}][{selected_command.description}]🔐 confirmed.")
            else:
                logger.info(f"[{OneColors.DARK_RED}][{selected_command.description}]❌ cancelled.")
            return confirm_answer
        return True

    def _create_context(self, selected_command: Command) -> ExecutionContext:
        """Creates a context dictionary for the selected command."""
        return ExecutionContext(
            name=selected_command.description,
            args=tuple(),
            kwargs={},
            action=selected_command,
        )

    async def _run_action_with_spinner(self, command: Command) -> Any:
        """Runs the action of the selected command with a spinner."""
        with self.console.status(
            command.spinner_message,
            spinner=command.spinner_type,
            spinner_style=command.spinner_style,
            **command.spinner_kwargs,
        ):
            return await command()

    async def _handle_action_error(self, selected_command: Command, error: Exception) -> bool:
        """Handles errors that occur during the action of the selected command."""
        logger.exception(f"Error executing '{selected_command.description}': {error}")
        self.console.print(f"[{OneColors.DARK_RED}]An error occurred while executing "
                           f"{selected_command.description}:[/] {error}")
        if self.confirm_on_error and not self._never_confirm:
            return await async_confirm("An error occurred. Do you wish to continue?")
        if self._never_confirm:
            return True
        return False 

    async def process_command(self) -> bool:
        """Processes the action of the selected command."""
        choice = await self.session.prompt_async()
        selected_command = self.get_command(choice)
        if not selected_command:
            logger.info(f"[{OneColors.LIGHT_YELLOW}] Invalid command '{choice}'.")
            return True
        self.last_run_command = selected_command

        if selected_command == self.exit_command:
            logger.info(f"🔙 Back selected: exiting {self.get_title()}")
            return False

        if not await self._should_run_action(selected_command):
            logger.info(f"[{OneColors.DARK_RED}] {selected_command.description} cancelled.")
            return True

        context = self._create_context(selected_command)
        context.start_timer()
        try:
            await self.hooks.trigger(HookType.BEFORE, context)

            if selected_command.spinner:
                result = await self._run_action_with_spinner(selected_command)
            else:
                result = await selected_command()
            context.result = result
            await self.hooks.trigger(HookType.ON_SUCCESS, context)
        except Exception as error:
            context.exception = error
            await self.hooks.trigger(HookType.ON_ERROR, context)
            if not context.exception:
                logger.info(f"✅ Recovery hook handled error for '{selected_command.description}'")
                context.result = result
            else:
                return await self._handle_action_error(selected_command, error)
        finally:
            context.stop_timer()
            await self.hooks.trigger(HookType.AFTER, context)
            await self.hooks.trigger(HookType.ON_TEARDOWN, context)
        return True

    async def headless(self, command_key: str, return_context: bool = False) -> Any:
        """Runs the action of the selected command without displaying the menu."""
        self.debug_hooks()
        selected_command = self.get_command(command_key)
        self.last_run_command = selected_command

        if not selected_command:
            logger.info("[Headless] Back command selected. Exiting menu.")
            return

        logger.info(f"[Headless] 🚀 Running: '{selected_command.description}'")

        if not await self._should_run_action(selected_command):
            raise FalyxError(f"[Headless] '{selected_command.description}' cancelled by confirmation.")

        context = self._create_context(selected_command)
        context.start_timer()
        try:
            await self.hooks.trigger(HookType.BEFORE, context)

            if selected_command.spinner:
                result = await self._run_action_with_spinner(selected_command)
            else:
                result = await selected_command()
            context.result = result

            await self.hooks.trigger(HookType.ON_SUCCESS, context)
            logger.info(f"[Headless] ✅ '{selected_command.description}' complete.")
        except (KeyboardInterrupt, EOFError):
            raise FalyxError(f"[Headless] ⚠️ '{selected_command.description}' interrupted by user.")
        except Exception as error:
            context.exception = error
            await self.hooks.trigger(HookType.ON_ERROR, context)
            if not context.exception:
                logger.info(f"[Headless] ✅ Recovery hook handled error for '{selected_command.description}'")
                return True
            raise FalyxError(f"[Headless] ❌ '{selected_command.description}' failed.") from error
        finally:
            context.stop_timer()
            await self.hooks.trigger(HookType.AFTER, context)
            await self.hooks.trigger(HookType.ON_TEARDOWN, context)

        return context if return_context else context.result

    def _set_retry_policy(self, selected_command: Command) -> None:
        """Sets the retry policy for the command based on CLI arguments."""
        assert isinstance(self.cli_args, Namespace), "CLI arguments must be provided."
        if self.cli_args.retries or self.cli_args.retry_delay or self.cli_args.retry_backoff:
            selected_command.retry_policy.enabled = True
            if self.cli_args.retries:
                selected_command.retry_policy.max_retries = self.cli_args.retries
            if self.cli_args.retry_delay:
                selected_command.retry_policy.delay = self.cli_args.retry_delay
            if self.cli_args.retry_backoff:
                selected_command.retry_policy.backoff = self.cli_args.retry_backoff
            selected_command.update_retry_policy(selected_command.retry_policy)

    def print_message(self, message: str | Markdown | dict[str, Any]) -> None:
        """Prints a message to the console."""
        if isinstance(message, (str, Markdown)):
            self.console.print(message)
        elif isinstance(message, dict):
            self.console.print(
                *message.get("args", tuple()),
                **message.get("kwargs", {}),
            )
        else:
            raise TypeError(
                "Message must be a string, Markdown, or dictionary with args and kwargs."
            )

    async def menu(self) -> None:
        """Runs the menu and handles user input."""
        logger.info(f"Running menu: {self.get_title()}")
        self.debug_hooks()
        if self.welcome_message:
            self.print_message(self.welcome_message)
        while True:
            self.console.print(self.table)
            try:
                task = asyncio.create_task(self.process_command())
                should_continue = await task
                if not should_continue:
                    break
            except (EOFError, KeyboardInterrupt):
                logger.info(f"[{OneColors.DARK_RED}]EOF or KeyboardInterrupt. Exiting menu.")
                break
        logger.info(f"Exiting menu: {self.get_title()}")
        if self.exit_message:
            self.print_message(self.exit_message)

    async def run(self, parsers: FalyxParsers | None = None) -> None:
        """Run Falyx CLI with structured subcommands."""
        parsers = parsers or get_arg_parsers()
        self.cli_args = parsers.root.parse_args()

        if self.cli_args.verbose:
            logging.getLogger("falyx").setLevel(logging.DEBUG)

        if self.cli_args.debug_hooks:
            logger.debug("✅ Enabling global debug hooks for all commands")
            self.register_all_with_debug_hooks()

        if self.cli_args.command == "list":
            await self._show_help()
            sys.exit(0)

        if self.cli_args.command == "version" or self.cli_args.version:
            self.console.print(f"[{OneColors.GREEN_b}]Falyx CLI v{__version__}[/]")
            sys.exit(0)

        if self.cli_args.command == "preview":
            command = self.get_command(self.cli_args.name)
            if not command:
                self.console.print(f"[{OneColors.DARK_RED}]❌ Command '{self.cli_args.name}' not found.[/]")
                sys.exit(1)
            self.console.print(f"Preview of command '{command.key}': {command.description}")
            await command.preview()
            sys.exit(0)

        if self.cli_args.command == "run":
            command = self.get_command(self.cli_args.name)
            if not command:
                self.console.print(f"[{OneColors.DARK_RED}]❌ Command '{self.cli_args.name}' not found.[/]")
                sys.exit(1)
            self._set_retry_policy(command)
            try:
                await self.headless(self.cli_args.name)
            except FalyxError as error:
                self.console.print(f"[{OneColors.DARK_RED}]❌ Error: {error}[/]")
                sys.exit(1)
            sys.exit(0)

        if self.cli_args.command == "run-all":
            matching = [
                cmd for cmd in self.commands.values()
                if self.cli_args.tag.lower() in (tag.lower() for tag in cmd.tags)
            ]
            if not matching:
                self.console.print(f"[{OneColors.LIGHT_YELLOW}]⚠️ No commands found with tag: '{self.cli_args.tag}'[/]")
                sys.exit(1)

            self.console.print(f"[{OneColors.CYAN_b}]🚀 Running all commands with tag:[/] {self.cli_args.tag}")
            for cmd in matching:
                self._set_retry_policy(cmd)
                await self.headless(cmd.key)
            sys.exit(0)

        await self.menu()



"""command.py
Any Action or Command is callable and supports the signature:
    result = thing(*args, **kwargs)

This guarantees:
- Hook lifecycle (before/after/error/teardown)
- Timing
- Consistent return values
"""
from __future__ import annotations

from typing import Any, Callable

from prompt_toolkit.formatted_text import FormattedText
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr, field_validator
from rich.console import Console
from rich.tree import Tree

from falyx.action import Action, BaseAction
from falyx.context import ExecutionContext
from falyx.debug import register_debug_hooks
from falyx.execution_registry import ExecutionRegistry as er
from falyx.hook_manager import HookManager, HookType
from falyx.retry import RetryPolicy
from falyx.themes.colors import OneColors
from falyx.utils import _noop, ensure_async, logger

console = Console()


class Command(BaseModel):
    """Class representing an command in the menu."""
    key: str
    description: str
    aliases: list[str] = Field(default_factory=list)
    action: BaseAction | Callable[[], Any] = _noop
    args: tuple = ()
    kwargs: dict[str, Any] = Field(default_factory=dict)
    help_text: str = ""
    color: str = OneColors.WHITE
    confirm: bool = False
    confirm_message: str = "Are you sure?"
    preview_before_confirm: bool = True
    spinner: bool = False
    spinner_message: str = "Processing..."
    spinner_type: str = "dots"
    spinner_style: str = OneColors.CYAN
    spinner_kwargs: dict[str, Any] = Field(default_factory=dict)
    hooks: "HookManager" = Field(default_factory=HookManager)
    retry: bool = False
    retry_all: bool = False
    retry_policy: RetryPolicy = Field(default_factory=RetryPolicy)
    tags: list[str] = Field(default_factory=list)
    logging_hooks: bool = False

    _context: ExecutionContext | None = PrivateAttr(default=None)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def model_post_init(self, __context: Any) -> None:
        """Post-initialization to set up the action and hooks."""
        if self.retry and isinstance(self.action, Action):
            self.action.enable_retry()
        elif self.retry_policy and isinstance(self.action, Action):
            self.action.set_retry_policy(self.retry_policy)
        elif self.retry:
            logger.warning(f"[Command:{self.key}] Retry requested, but action is not an Action instance.")
        if self.retry_all:
            self.action.enable_retries_recursively(self.action, self.retry_policy)

        if self.logging_hooks and isinstance(self.action, BaseAction):
            register_debug_hooks(self.action.hooks)

    @field_validator("action", mode="before")
    @classmethod
    def wrap_callable_as_async(cls, action: Any) -> Any:
        if isinstance(action, BaseAction):
            return action
        elif callable(action):
            return ensure_async(action)
        raise TypeError("Action must be a callable or an instance of BaseAction")

    def __str__(self):
        return f"Command(key='{self.key}', description='{self.description}')"

    async def __call__(self, *args, **kwargs):
        """Run the action with full hook lifecycle, timing, and error handling."""
        combined_args = args + self.args
        combined_kwargs = {**self.kwargs, **kwargs}
        context = ExecutionContext(
            name=self.description,
            args=combined_args,
            kwargs=combined_kwargs,
            action=self,
        )
        self._context = context
        context.start_timer()

        try:
            await self.hooks.trigger(HookType.BEFORE, context)
            result = await self.action(*combined_args, **combined_kwargs)
            context.result = result
            await self.hooks.trigger(HookType.ON_SUCCESS, context)
            return context.result
        except Exception as error:
            context.exception = error
            await self.hooks.trigger(HookType.ON_ERROR, context)
            if context.result is not None:
                logger.info(f"✅ Recovered: {self.key}")
                return context.result
            raise error
        finally:
            context.stop_timer()
            await self.hooks.trigger(HookType.AFTER, context)
            await self.hooks.trigger(HookType.ON_TEARDOWN, context)
            er.record(context)

    @property
    def result(self) -> Any:
        """Get the result of the action."""
        return self._context.result if self._context else None

    @property
    def confirmation_prompt(self) -> FormattedText:
        """Generate a styled prompt_toolkit FormattedText confirmation message."""
        if self.confirm_message and self.confirm_message != "Are you sure?":
            return FormattedText([
                ("class:confirm", self.confirm_message)
            ])

        action_name = getattr(self.action, "__name__", None)
        if isinstance(self.action, BaseAction):
            action_name = self.action.name

        prompt = [(OneColors.WHITE, "Confirm execution of ")]

        prompt.append((OneColors.BLUE_b, f"{self.key}"))
        prompt.append((OneColors.BLUE_b, f" — {self.description} "))

        if action_name:
            prompt.append(("class:confirm", f"(calls `{action_name}`) "))

        if self.args or self.kwargs:
            prompt.append((OneColors.DARK_YELLOW, f"with args={self.args}, kwargs={self.kwargs} "))

        return FormattedText(prompt)

    def log_summary(self):
        if self._context:
            self._context.log_summary()

    async def preview(self):
        label = f"[{OneColors.GREEN_b}]Command:[/] '{self.key}' — {self.description}"

        if hasattr(self.action, "preview") and callable(self.action.preview):
            tree = Tree(label)
            await self.action.preview(parent=tree)
            console.print(tree)
        elif callable(self.action):
            console.print(f"{label}")
            console.print(
                f"[{OneColors.LIGHT_RED_b}]→ Would call:[/] {self.action.__name__} "
                f"[dim](args={self.args}, kwargs={self.kwargs})[/dim]"
            )
        else:
            console.print(f"{label}")
            console.print(f"[{OneColors.DARK_RED}]⚠️ Action is not callable or lacks a preview method.[/]")



# falyx/__main__.py

import asyncio
import logging

from falyx.action import Action
from falyx.falyx import Falyx


def build_falyx() -> Falyx:
    """Build and return a Falyx instance with all your commands."""
    app = Falyx(title="🚀 Falyx CLI")

    # Example commands
    app.add_command(
        key="B",
        description="Build project",
        action=Action("Build", lambda: print("📦 Building...")),
        tags=["build"]
    )

    app.add_command(
        key="T",
        description="Run tests",
        action=Action("Test", lambda: print("🧪 Running tests...")),
        tags=["test"]
    )

    app.add_command(
        key="D",
        description="Deploy project",
        action=Action("Deploy", lambda: print("🚀 Deploying...")),
        tags=["deploy"]
    )

    return app

if __name__ == "__main__":
    logging.basicConfig(level=logging.WARNING)
    falyx = build_falyx()
    asyncio.run(falyx.cli())



"""hook_manager.py"""
from __future__ import annotations

import inspect
from enum import Enum
from typing import Awaitable, Callable, Dict, List, Optional, Union

from falyx.context import ExecutionContext
from falyx.utils import logger

Hook = Union[
    Callable[[ExecutionContext], None],
    Callable[[ExecutionContext], Awaitable[None]]
]


class HookType(Enum):
    """Enum for hook types to categorize the hooks."""
    BEFORE = "before"
    ON_SUCCESS = "on_success"
    ON_ERROR = "on_error"
    AFTER = "after"
    ON_TEARDOWN = "on_teardown"

    @classmethod
    def choices(cls) -> List[HookType]:
        """Return a list of all hook type choices."""
        return list(cls)

    def __str__(self) -> str:
        """Return the string representation of the hook type."""
        return self.value


class HookManager:
    def __init__(self) -> None:
        self._hooks: Dict[HookType, List[Hook]] = {
            hook_type: [] for hook_type in HookType
        }

    def register(self, hook_type: HookType, hook: Hook):
        if hook_type not in HookType:
            raise ValueError(f"Unsupported hook type: {hook_type}")
        self._hooks[hook_type].append(hook)

    def clear(self, hook_type: Optional[HookType] = None):
        if hook_type:
            self._hooks[hook_type] = []
        else:
            for ht in self._hooks:
                self._hooks[ht] = []

    async def trigger(self, hook_type: HookType, context: ExecutionContext):
        if hook_type not in self._hooks:
            raise ValueError(f"Unsupported hook type: {hook_type}")
        for hook in self._hooks[hook_type]:
            try:
                if inspect.iscoroutinefunction(hook):
                    await hook(context)
                else:
                    hook(context)
            except Exception as hook_error:
                logger.warning(f"⚠️ Hook '{hook.__name__}' raised an exception during '{hook_type}'"
                               f" for '{context.name}': {hook_error}")

                if hook_type == HookType.ON_ERROR:
                    assert isinstance(context.exception, BaseException)
                    raise context.exception from hook_error



import logging

from .action import Action, ActionGroup, ChainedAction, ProcessAction
from .command import Command
from .context import ExecutionContext, ResultsContext
from .execution_registry import ExecutionRegistry
from .falyx import Falyx

logger = logging.getLogger("falyx")

__version__ = "0.1.0"

__all__ = [
    "Action",
    "ChainedAction",
    "ActionGroup",
    "ProcessAction",
    "Falyx",
    "Command",
    "ExecutionContext",
    "ResultsContext",
    "ExecutionRegistry",
]
